#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef USE_LIBEDIT
#include <editline/readline.h>
#endif
#ifdef USE_EDITLINE
#include <editline.h>
#endif
#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif
#ifdef USE_NLLINE
#include <nlline.h>
#endif
#if defined(USE_GETLINE) || defined(USE_FGETS)
#include "readline.h"
#define readline(prompt) nlsh_readline(prompt)
#define add_history(line) nlsh_add_history(line)
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "args.h"
#include "val.h"

static int sigint_handled = 0;
static int sigchld_handled = 0;

#if 0
static void signal_reset(void)
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_DFL;
  sigaction(SIGINT, &sa, NULL);

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_DFL;
  sigaction(SIGCHLD, &sa, NULL);
}
#endif

static int exec_command(char *argv[], char *envp[], char *path[])
{
  int n, i, flags, fd, out, save[3];
  char **args = NULL;
  char *outstr = NULL, *command = NULL;
  enum {
    EXEC_ERR_UNKNOWN = 0,
    EXEC_ERR_NOFILE,
    EXEC_ERR_FILEERR,
  } e = EXEC_ERR_UNKNOWN;

  for (i = 0; i < 3; i++)
    save[i] = -1;

  while (1) {
    n = search_argv(argv, "<");
    if (n < 0)
      break;
    args = extract_argv(argv, n, 2);
    if (get_argc(args) < 2) {
      e = EXEC_ERR_NOFILE;
      goto err;
    }
    fd = open(args[1], O_RDONLY);
    if (fd < 0) {
      e = EXEC_ERR_FILEERR;
      goto err;
    }
    if (save[0] < 0)
      save[0] = dup(0);
    close(0);
    dup2(fd, 0);
    close(fd);
    free_argv(args);
    args = NULL;
  }

  while (1) {
    n = search_argv(argv, ">");
    if (n < 0) n = search_argv(argv, "2>");
    if (n < 0) n = search_argv(argv, ">>");
    if (n < 0) n = search_argv(argv, "2>>");
    if (n < 0)
      break;
    args = extract_argv(argv, n, 2);
    if (get_argc(args) < 2) {
      e = EXEC_ERR_NOFILE;
      goto err;
    }
    if (!strcmp(args[0], ">>") || !strcmp(args[0], "2>>")) {
      flags = O_WRONLY | O_CREAT | O_APPEND;
    } else {
      flags = O_WRONLY | O_CREAT | O_TRUNC;
    }
    if (!strcmp(args[0], "2>") || !strcmp(args[0], "2>>")) {
      out = 2;
    } else {
      out = 1;
    }

    fd = open(args[1], flags, 0644);
    if (fd < 0) {
      e = EXEC_ERR_FILEERR;
      goto err;
    }
    if (save[out] < 0)
      save[out] = dup(out);
    close(out);
    dup2(fd, out);
    close(fd);
    free_argv(args);
    args = NULL;
  }

  do {
    n = search_argv(argv, "1>&2");
    if (n < 0) n = search_argv(argv, "2>&1");
    if (n < 0)
      break;
    args = extract_argv(argv, n, 1);
    if (!strcmp(args[0], "1>&2")) {
      if (save[1] < 0)
	save[1] = dup(1);
      close(1);
      dup2(2, 1);
    } else {
      if (save[2] < 0)
	save[2] = dup(2);
      close(2);
      dup2(1, 2);
    }
    free_argv(args);
    args = NULL;
  } while (n >= 0);

  if (get_argc(argv) == 0) {
    goto err;
  }

  if ((strchr("./", argv[0][0]) == NULL) && (path != NULL)) {
    for (i = 0; path[i]; i++) {
      command = malloc(strlen(path[i]) + strlen(argv[0]) + 2);
      if (command == NULL)
	goto err;
      sprintf(command, "%s/%s", path[i], argv[0]);
      fd = open(command, O_RDONLY);
      if (fd >= 0) {
	close(fd);
	break;
      }
      free(command);
      command = NULL;
    }
  }
  if (command == NULL)
    command = strdup(argv[0]);
  if (command == NULL)
    goto err;

#if 0
  signal_reset();
#endif

  execve(command, argv, envp);

err:
  if (args)
    free_argv(args);

  if (command != NULL)
    free(command);

  for (i = 0; i < 3; i++) {
    if (save[i] >= 0) {
      dup2(save[i], i);
      close(save[i]);
    }
  }

  switch (e) {
  case EXEC_ERR_NOFILE:  outstr = "No file selected.\n"; break;
  case EXEC_ERR_FILEERR: outstr = "File not found.\n"; break;
  case EXEC_ERR_UNKNOWN:
  default:
    outstr = "Unknown error.\n";
    break;
  }
  if (outstr)
    write(2, outstr, strlen(outstr));

  return -1;
}

static pid_t exec_argv(char *exec[], char *argv[], char *envp[], char *path[])
{
  int n, r, done, status;
  pid_t pid, first_child = 0;
  int fildes[2], bg = 0, ischild = 0, haschild = 0, pipe_in = -1;
  char *outstr;

  while (get_argc(argv) > 0) {
    if (!strcmp(argv[0], "(")) {
      delete_argv(argv, 0, 1);
      n = rsearch_argv(argv, ")");
      if (n < 0) {
	exec = extract_argv(argv, 0, -1);
      } else {
	delete_argv(argv, n, 1);
	for (; argv[n]; n++) {
	  if (!strcmp(argv[n], "|"))
	    break;
	}
	if (argv[n] == NULL) {
	  exec = extract_argv(argv, 0, -1);
	} else {
	  exec = extract_argv(argv, 0, n);
	  delete_argv(argv, 0, 1);
	}
      }
    } else {
      n = search_argv(argv, "|");
      if (n < 0) {
	exec = extract_argv(argv, 0, -1);
      } else {
	exec = extract_argv(argv, 0, n);
	delete_argv(argv, 0, 1);
      }
    }

    if (get_argc(exec) == 0)
      break;

    n = rsearch_argv(exec, "&");
    if (n >= 0) {
      delete_argv(exec, n, 1);
      bg = 1;
    }

    fildes[0] = -1;
    fildes[1] = -1;

    if (get_argc(argv) > 0) {
      if (pipe2(fildes, 0) < 0)
	break;
    }

    pid = fork();
    if (pid < 0)
      break;
    if (pid == 0) {
      /* child process */

      ischild = 1;
      haschild = 0;

      setpgid(0, first_child);

      free_argv(argv);
      argv = NULL;

      if (pipe_in >= 0) {
	close(0);
	dup2(pipe_in, 0);
	close(pipe_in);
	pipe_in = -1;
      }
      if (fildes[0] >= 0) {
	close(fildes[0]);
      }
      if (fildes[1] >= 0) {
	close(1);
	dup2(fildes[1], 1);
	close(fildes[1]);
      }

      if ((search_argv(exec, "|") >= 0) || (search_argv(exec, "(") >= 0)) {
	argv = exec;
	exec = NULL;
	continue;
      }

      exec_command(exec, envp, path);
      break;
    }

    /* parent process */

    haschild = 1;

    if (first_child == 0)
      first_child = pid;

    if (pipe_in >= 0) {
      close(pipe_in);
      pipe_in = -1;
    }
    if (fildes[0] >= 0)
      pipe_in = fildes[0];
    if (fildes[1] >= 0)
      close(fildes[1]);

    free_argv(exec);
    exec = NULL;
  }

  if (pipe_in >= 0)
    close(pipe_in);

  free_argv(exec);
  free_argv(argv);

  if (haschild && !bg) {
    done = -1;
    while (done <= 0) {
      sigint_handled = 0;
      r = wait4(-first_child, &status, 0, NULL);
      outstr = "Unknown.\n";
      if (r < 0) {
	if (errno == ECHILD) {
	  if (done < 0) {
	    /* Wait for child to start and setpgid() */
	  } else if (done == 0) {
	    done = 1;
	  }
	  outstr = "ECHILD.\n";
	}
	if (errno == EINTR) {
	  outstr = "EINTR.\n";
	  if (sigint_handled) {
	    if (first_child > 0)
	      killpg(first_child, SIGINT);
	  }
	}
      } else {
	if (WIFSTOPPED(status)) {
	  outstr = "Stopped.\n";
	} else if (WIFEXITED(status)) {
	  outstr = "Exited.\n";
	} else if (WIFCONTINUED(status)) {
	  outstr = "Cont.\n";
	} else if (WIFSIGNALED(status)) {
	  switch (WTERMSIG(status)) {
	  case SIGINT:
	    outstr = "Received signal. (SIGINT)\n";
	    break;
	  case SIGTERM:
	    outstr = "Received signal. (SIGTERM)\n";
	    break;
	  default:
	    outstr = "Received signal. (unknown)\n";
	    break;
	  }
	}
	done = 0;
      }
#if 1
      outstr = NULL;
#endif
      if (!ischild && (outstr != NULL))
	write(2, outstr, strlen(outstr));
    }
  }

  if (ischild) {
    _exit(0);
  }

  return first_child;
}

static int ismatch_wildcard(char *wildcard, char *filename)
{
  int i, j;

  for (i = 0;; i++) {
    switch (wildcard[i]) {
    case '*':
      for (j = 0;; j++) {
	if (ismatch_wildcard(wildcard + i + 1, filename + i + j))
	  return 1;
	if (filename[i + j] == '\0')
	  break;
      }
      return 0;
    case '?':
      break;
    default:
      if (wildcard[i] != filename[i])
	return 0;
      break;
    }
    if (wildcard[i] == '\0')
      break;
  }

  return 1;
}

static char **_expand_wildcard(char *argv[], int pos,
			       char *dirname, char *pathname)
{
  char *filename = NULL, *p;
  DIR *dirp = NULL;
  struct dirent *dirent;

  if (dirname == NULL) {
    if (pathname[0] == '/') {
      dirname = "/";
      pathname++;
    } else
      dirname = "";
  }

  filename = strdup(pathname);
  if (filename == NULL)
    goto ret;
  pathname = strchr(filename, '/');
  if (pathname != NULL) {
    *pathname = '\0';
    pathname++;
  }

  dirp = opendir(dirname[0] ? dirname : ".");
  if (dirp == NULL)
    goto ret;

  while (1) {
    dirent = readdir(dirp);
    if (dirent == NULL)
      break;
    if ((!strcmp(dirent->d_name, ".") || !strcmp(dirent->d_name, "..")) &&
	(filename[0] == '*'))
      continue;
    if (ismatch_wildcard(filename, dirent->d_name)) {
      p = malloc(strlen(dirname) + 1 + strlen(dirent->d_name) + 1);
      if (p == NULL)
	goto ret;
      sprintf(p, "%s%s%s", dirname, dirname[0] ? "/" : "", dirent->d_name);
      if (pathname != NULL) {
	argv = _expand_wildcard(argv, pos, p, pathname);
      } else {
	argv = insert_argv(argv, pos, p);
      }
      free(p);
    }
  }

ret:
  if (filename != NULL)
    free(filename);
  if (dirp)
    closedir(dirp);

  return argv;
}

static char **expand_wildcard(char *argv[])
{
  int argc, i;

  argc = get_argc(argv);

  for (i = argc - 1; i > 0; i--) {
    if (!strchr(argv[i], '*') && !strchr(argv[i], '?'))
      continue;
    argv = _expand_wildcard(argv, i + 1, NULL, argv[i]);
    delete_argv(argv, i, 1);
  }

  return argv;
}

static int exec_line(char *line, char *envp[], char *path)
{
  char **argv, **path_argv;
  pid_t pid;

  argv = make_argv(line);

  argv = expand_wildcard(argv);

  if ((get_argc(argv) > 1) && !strcmp(argv[0], "cd")) {
    if (chdir(argv[1]) < 0) {
      write(2, argv[1], strlen(argv[1]));
      write(2, ": No such file or directory.\n", 29);
      return -1;
    }
    return 0;
  }

  path_argv = make_argv_punc(path);
  pid = exec_argv(NULL, argv, envp, path_argv);
  free_argv(path_argv);

  if (pid == 0)
    return -1;

  return pid;
}

void sigint_handler(int arg)
{
  sigint_handled++;
  return;
}

void sigchld_handler(int arg)
{
  sigchld_handled++;
  return;
}

static void signal_init(void)
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sigint_handler;
  sigaction(SIGINT, &sa, NULL);

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sigchld_handler;
  sigaction(SIGCHLD, &sa, NULL);
}

static void init(void)
{
  signal_init();
}

static char *make_prompt(void)
{
  char *p, *user, *host, *home;
  char cwd[256], hostname[256];
  static char *prompt = NULL;

  user = val_get("USER");
  if (user == NULL)
    user = val_get("LOGNAME");

  host = val_get("HOST");
  if (host == NULL)
    host = val_get("HOSTNAME");

  home = val_get("HOME");

  if (getcwd(cwd, sizeof(cwd)) == NULL)
    cwd[0] = 0;

  if (user == NULL)
    user = "user";

  if ((host != NULL) && (strlen(host) < sizeof(hostname))) {
    strcpy(hostname, host);
    p = strchr(hostname, '.');
    if (p)
      *p = '\0';
    host = hostname;
  } else {
    host = "host";
  }

  if ((home != NULL) && (strlen(home) > 1) &&
      !strncmp(cwd, home, strlen(home))) {
    memmove(cwd + 1, cwd + strlen(home), strlen(cwd) - strlen(home) + 1);
    cwd[0] = '~';
  }

  prompt = realloc(prompt, 32 + strlen(user) + strlen(host) + strlen(cwd));
  if (prompt != NULL) {
    sprintf(prompt, "%s@%s:%s>%% ", user, host, cwd);
  }

  return prompt ? prompt : "> ";
}

static char **file_search(char *pathname, char *dirname, char *filename,
			  char **argv)
{
  DIR *dirp = NULL;
  char *p;
  struct dirent *dirent;

  dirp = opendir(dirname[0] ? dirname : ".");
  if (dirp == NULL)
    goto err;

  while (1) {
    dirent = readdir(dirp);
    if (dirent == NULL)
      break;
    if (!strcmp(dirent->d_name, ".") || !strcmp(dirent->d_name, ".."))
      continue;
    if (!strncmp(filename, dirent->d_name, strlen(filename))) {
      if (dirent->d_type == DT_DIR) {
	p = malloc(strlen(dirent->d_name) + 2);
	if (p != NULL) {
	  strcpy(p, dirent->d_name);
	  strcat(p, "/");
	  argv = add_argv(argv, p);
	  free(p);
	} else {
	  argv = add_argv(argv, dirent->d_name);
	}
      } else {
	argv = add_argv(argv, dirent->d_name);
      }
      if (argv == NULL)
	goto err;
    }
  }

  goto ret;

err:
  if (argv != NULL) {
    free_argv(argv);
    argv = NULL;
  }

ret:
  if (dirp != NULL)
    closedir(dirp);

  return argv;
}

char **readline_completion_func(const char *line, int start, int end)
{
  char *pathname = NULL, *dirname = NULL, *filename, *complete = NULL;
  char **argv = NULL, **r = NULL, *p;
  int i, n;

  pathname = malloc(end - start + 1);
  if (pathname == NULL)
    goto ret;

  strncpy(pathname, line, end - start);
  pathname[end - start] = '\0';

  p = strrchr(pathname, '/');
  if (p != NULL) {
    dirname = malloc(p - pathname + 1);
    if (dirname == NULL)
      goto ret;
    strncpy(dirname, pathname, p - pathname);
    dirname[p - pathname] = '\0';
    filename = p + 1;
  } else {
    dirname = strdup("");
    if (dirname == NULL)
      goto ret;
    filename = pathname;
  }

  argv = file_search(pathname, dirname, filename, NULL);

  if ((argv == NULL) || (argv[0] == NULL))
    goto ret;

  complete = malloc(strlen(dirname) + 1 + strlen(argv[0]) + 1);
  if (complete == NULL)
    goto ret;
  strcpy(complete, argv[0]);

  for (n = 0; argv[n]; n++) {
    for (i = 0; complete[i]; i++) {
      if (argv[n][i] != complete[i]) {
	complete[i] = '\0';
	break;
      }
    }
  }

  if (dirname[0]) {
    memmove(complete + strlen(dirname) + 1, complete, strlen(complete) + 1);
    strcpy(complete, dirname);
    complete[strlen(dirname)] = '/';
  }

  argv = insert_argv(argv, 0, pathname);
  if (strcmp(pathname, complete) != 0) {
    argv = insert_argv(argv, 0, complete);
  }

ret:
  if (pathname != NULL)
    free(pathname);
  if (dirname != NULL)
    free(dirname);
  if (complete != NULL)
    free(complete);

  if (argv != NULL) {
    n = get_argc(argv);
    r = malloc(sizeof(*r) * (n + 1));
    if (r != NULL) {
      for (i = 0; i < n; i++) {
	r[i] = strdup(argv[i]);
      }
      r[n] = NULL;
    }
    free_argv(argv);
  }

  return r;
}

int main(int argc, char *argv[], char *envp[])
{
  char *line, *outstr, *prompt;
  int i, finished = 0, status;
  pid_t pid;

#if 0
  for (i = 0; envp[i]; i++)
    printf("envp[%d]: %s\n", i, envp[i]);
  for (i = 0; i < argc; i++)
    printf("argv[%d]: %s\n", i, argv[i]);
#endif

  init();

  for (i = 0; envp[i]; i++) {
    val_add_from_string(envp[i]);
  }

  rl_attempted_completion_function = readline_completion_func;

  while (!finished) {
    prompt = make_prompt();

    line = readline(prompt);
    if (line == NULL) {
      write(1, "\n", 1);
      while (1) {
	pid = wait4(-1, &status, WNOHANG, NULL);
	if (pid <= 0)
	  break;
	outstr = "Unknown.\n";
	if (WIFSTOPPED(status))        outstr = "Stopped.\n";
	else if (WIFEXITED(status))    outstr = "Exited.\n";
	else if (WIFCONTINUED(status)) outstr = "Cont.\n";
	else if (WIFSIGNALED(status))  outstr = "Received signal.\n";
	if (outstr)
	  printf("%s [%d]\n", outstr, pid);
      }
      continue;
    }

    if (!strcmp(line, "exit"))
      finished = 1;
    else if (exec_line(line, envp, val_get("PATH")) >= 0)
      add_history(line);

    free(line);
  }

  return 0;
}
